/*
	"main.c" - Sample Code

	Copyright (C) 1998  Sony Computer Entertainment Europe
	All rights Reserved
*/

// =======================================================================
// PLEASE SEE THE HTML FILE "DINODEMO.HTM" FOR MORE DETAILED DOCUMENTATION
// =======================================================================

	// The entry point and 'main loop'. There are two 'main loops' (one for
	// Intro Mode and one for Main Mode). The difference is that one follows
	// a set camera path with no user interaction, and one reacts to user
	// input.

#define _MAIN_

#include <sys/types.h>
#include <libmath.h>
#include <libetc.h>
#include <libgte.h>
#include <libgpu.h>
#include <libgs.h>
#include <libhmd.h>
#include <rand.h>

#include "3d.h"
#include "main.h"
#include "display.h"
#include "water.h"
#include "pad.h"
#include "music/xmplayer.h"

#ifdef DEVKIT
//#define MESSAGES_ON
//#define ARTIST
#endif

#ifdef DEVKIT
#include <libsn.h>
#else
#include <libcd.h>
#endif

#ifndef DEVKIT
unsigned long _ramsize   = 0x00200000;
unsigned long _stacksize = 0x00010000;
#endif

extern char BACKOFF_RAW[];

static const int zero = 0;
// ===================================================================
// ===================================================================
// ===================================================================
extern SVECTOR ReflectionCamera;

typedef enum { FADE_UP, WAITING, RIPPLE1, RIPPLE2, DINO_MOVE, PAN_OUT } INIT_PART;

// ===================================================================
// ====  I N T R O   S E Q U E N C E =================================
// ===================================================================


		// Sequence seen at the start (not under user control). The
		// scene fades in, the dinosaur appears, backs off and moves
		// into the position ready for interaction.


bool inline introSequence(void) {
	u_long pad;

			// Primitives used for fading in the scene
	DR_MODE semiTransMinusMode,normalMode;
	POLY_F4	fadePoly;
	u_long fadeValue = 256;

	bool
		quit = FALSE,
		quitIntro = FALSE;

	int
		frameNum = 0,
		waitTime = 0,
		distanceToMove = 0,
		finalTTopsZ;

	const int zero = 0;

		// The current 'stage' we are at during the intro. When we are
		// waiting for a timer, the state will switch to 'partAfterWait'
		// when the time elapses.
	INIT_PART
		initPart = FADE_UP,
		partAfterWait = WAITING;

	const int NUM_PAN_OUT_FRAMES = 512; // Number of frames we will dolly out over.

		// Hack - this area of VRAM contains part of the pond CLUT which is
		// black semi-transparent. Later on, we clear it (make it black
		// with no semi-trans which speeds things up quite considerably
		// when we switch models because there isn't a large group of
		// giant semitrans polygons being drawn over the screen.
	RECT rect = { 880+15,426,1,1 };

			// Initialise fade in semitrans polygon
	SetDrawMode(&semiTransMinusMode,TRUE,TRUE, GetTPage(2,2,0,0), 0);
	SetDrawMode(&normalMode,TRUE,TRUE, GetTPage(2,0,0,0), 0);
	SetPolyF4(&fadePoly);	// Polygon used to fade the screen in/out
	setXY4(&fadePoly,
		-SCREEN_WIDTH/2, -SCREEN_HEIGHT/2,
		 SCREEN_WIDTH/2, -SCREEN_HEIGHT/2,
		-SCREEN_WIDTH/2,  SCREEN_HEIGHT/2,
		 SCREEN_WIDTH/2,  SCREEN_HEIGHT/2);
	setRGB0(&fadePoly,fadeValue,fadeValue,fadeValue);
	setSemiTrans(&fadePoly,1);
	catPrim(&semiTransMinusMode,&fadePoly);
	catPrim(&fadePoly,&normalMode);
	termPrim(&normalMode);

		// Set the camera position to be right over the middle of the pond.
	CameraDistance = 1;
	CameraHeight = -700;

		// This call will install the appropriate rotations and translations
		// in the triceratops model, but not advance the frame.
	AnimateHMD(&triceratopsHMD, BACKOFF_RAW, FALSE);

		// Position the dinosaur
	finalTTopsZ = triceratopsHMD.position.vz;
	triceratopsHMD.position.vz = 10000;
	headMoveUpDown = headMoveLeftRight = 0;

		// Main loop
	while(!quit) {
		initNewDraw();
		pad = getPadButtons();

			// Actions depend on the current state of the intro.
		switch(initPart) {
			case FADE_UP:	// Fading in scene
				fadeValue-=4;
				if(fadeValue <= 0) {
					fadeValue = 0;

						// Wait for 200 frames after fading up.
					initPart = WAITING;
					partAfterWait = RIPPLE1;
					waitTime = 200;
				}
				setRGB0(&fadePoly,fadeValue,fadeValue,fadeValue);
				break;

			case WAITING:	// Waiting for the next state to begin
				if(waitTime > 0) {
					waitTime--;
				} else {
					initPart = partAfterWait;
				}
				break;

			case RIPPLE1:	// First ripple
				doStompSound();
				AddLump(20);
				doShock();

				waitTime = 50;	// Initialise wait
				partAfterWait = RIPPLE2;
				initPart = WAITING;
				break;

			case RIPPLE2:	// Second ripple
				doStompSound();
				AddLump(50);
				doShock();

				waitTime = 50;	// Initialise wait
				partAfterWait = DINO_MOVE;
				initPart = WAITING;
				distanceToMove = 1000;
				break;

			case DINO_MOVE:		// Move the dinosaur model towards the pond
					// Stomp 3 times as the dino moves forwards
				if(distanceToMove == 1000 || distanceToMove == 500 || distanceToMove == 100) {
					doStompSound();
					AddLump(20);
					doShock();
				}

					// The following complicated equation decelerates the dinosaur nicely,
					// rather than stopping it with a jolt as it reaches the right position.
				if(distanceToMove >= 0) {
					triceratopsHMD.position.vz = finalTTopsZ + 1000 - 1000 * rcos(1024 * distanceToMove / 1000)/ONE;
					distanceToMove -= 10;
				} else {
					waitTime = 100;	// Initialise wait
					partAfterWait = PAN_OUT;
					initPart = WAITING;

						// Setting this flag will cause the other environment
						// model to be used. The LoadImage will clear the CLUT
						// as described above.
					introModeOn = FALSE;
					LoadImage(&rect,(u_long*)&zero);
				}
				break;

			case PAN_OUT:
				frameNum++;

					// This equation smoothly moves the camera back.
				CameraDistance = FURTHEST_CAMERA_DISTANCE - FURTHEST_CAMERA_DISTANCE * (4096 + rcos(2048 * frameNum / NUM_PAN_OUT_FRAMES)) / (2*ONE);
				CameraHeight = -700 + 200 * rsin(2048 * frameNum / NUM_PAN_OUT_FRAMES) / ONE;
				CameraAngle--;

						// Do the roar (move head, open mouth, make sound)
				if(frameNum > 150 && frameNum < 300) {
					headMoveLeftRight = -300 * (4096 - rcos((frameNum-150) * 4096 / 150)) / (2*ONE);
					headMoveUpDown = -400 * (4096 - rcos((frameNum-150) * 4096 / 150)) / (2*ONE);
				}
				if(frameNum == 150) {
					mouthMoveFrame = 16;
					doRoarSound();
				}

					// Insert the appropriate translation/rotations, but do not advance frame.
				AnimateHMD(&triceratopsHMD, BACKOFF_RAW, FALSE);
				if(frameNum > (NUM_PAN_OUT_FRAMES - (118 * 2)) && (frameNum & 1)) {
					// Step the animation forward every other frame.
					AnimateHMD(&triceratopsHMD, BACKOFF_RAW, TRUE);
				}

				fadeInShadow();

				if(frameNum == NUM_PAN_OUT_FRAMES) {
					quit = TRUE;
				}
				break;
		}	// end switch

			// Calculate this so that it always looks at the pond.
		CameraLookUpDownAngle = ratan2(-CameraHeight - CAMERA_LEVEL_Y,CameraDistance);

		CameraRotate.vy = (-CameraAngle + ONE / 2);
		CameraRotate.vx = CameraLookUpDownAngle;
		CameraRotate.vz = 0;
		TopView.vpx = rsin(CameraAngle) * CameraDistance / ONE;
		TopView.vpz = rcos(CameraAngle) * CameraDistance / ONE;

		TopView.vpx += CameraShudder_LR;
		TopView.vpy = CameraHeight + CameraShudder_UD;

			// Adjust the sky's U/Vs so that it looks 3D.
		moveSky(TopView.vpx,TopView.vpz);

		processSounds();

		if(pad & PADRup) {
			quitIntro = TRUE;
			quit = TRUE;
		}

			// Draw the scene
		CreateOTforReflection();
		CreateOTforGround();
		CreateOTforMainScene();

		processController();

		completeDraw();
			// Draw the semi-trans 'fade-in' poly if necessary.
		if(fadeValue > 0) {
			DrawOTag((u_long *)&semiTransMinusMode);
		}

#ifdef MESSAGES_ON
		FntFlush(-1);
#endif
	}
		// return whether the user hit 'quit' during this intro.
	return quitIntro;
}

// ===================================================================
// ====  M A I N   L O O P ===========================================
// ===================================================================

#ifdef DEVKIT
int main(void) {
#else
int main(int argc, char *argv[]) {
#endif
	u_long
		pad,
		oldpad;

	u_char analogControls[4];

	int		// Dino head movement variables
		desiredHeadLeftRight,
		desiredHeadUpDown,
		oldDesiredHeadLeftRight = 0,
		oldDesiredHeadUpDown = 0,
		diff;
	bool quit = FALSE;

	init();

	quit = introSequence();

	while(!quit) {
		initNewDraw();
		pad = getPadButtons();
		getAnalogControllerValues(analogControls);

			// Adjust the camera position. Note that we always use the
			// analogControl array, even for a digital controller. This is
			// because the getAnalogControls function will 'fake' analog
			// values when no Dual Shock is inserted.

		CameraAngle -= ((int)analogControls[2] - 0x80)/3;
		CameraHeight += ((int)analogControls[3] - 0x80);

			// We take the difference between the old values and the new
			// because the analog values can 'twitch' between two consecutive
			// values very rapidly. By ignoring small changes we eliminate
			// the effect.

		desiredHeadLeftRight = ((int)analogControls[0] - 0x80) * 16;
		diff = oldDesiredHeadLeftRight - desiredHeadLeftRight;
		if(diff <= 16 && diff >= -16) {
			desiredHeadLeftRight = oldDesiredHeadLeftRight;
		} else {
			oldDesiredHeadLeftRight = desiredHeadLeftRight;
		}

		desiredHeadUpDown    = ((int)analogControls[1] - 0x80)* 12;
		diff = oldDesiredHeadUpDown - desiredHeadUpDown;
		if(diff <= 12 && diff >= -12) {
			desiredHeadUpDown = oldDesiredHeadUpDown;
		} else {
			oldDesiredHeadUpDown = desiredHeadUpDown;
		}

			// The analog control specifies a position for the head, not
			// a movement speed. The dino will move its head towards the
			// specified position rather than move straight to it.
#define HEADMOVESPEED 28

		if(headMoveLeftRight < desiredHeadLeftRight - HEADMOVESPEED) {
			headMoveLeftRight += HEADMOVESPEED;
		} else if(headMoveLeftRight <= desiredHeadLeftRight) {
			headMoveLeftRight = desiredHeadLeftRight;
		} else if(headMoveLeftRight > desiredHeadLeftRight + HEADMOVESPEED) {
			headMoveLeftRight -= HEADMOVESPEED;
		} else {
			headMoveLeftRight = desiredHeadLeftRight;
		}

		if(headMoveUpDown < (desiredHeadUpDown - HEADMOVESPEED)) {
			headMoveUpDown += HEADMOVESPEED;
		} else if(headMoveUpDown <= desiredHeadUpDown) {
			headMoveUpDown = desiredHeadUpDown;
		} else if(headMoveUpDown > (desiredHeadUpDown + HEADMOVESPEED)) {
			headMoveUpDown -= HEADMOVESPEED;
		} else {
			headMoveUpDown = desiredHeadUpDown;
		}

			// Clamp the camera heights and distances.

		if(CameraHeight > -LOWEST_CAMERA_Y) {
			CameraHeight = -LOWEST_CAMERA_Y;
		} else if(CameraHeight < -HIGHEST_CAMERA_Y) {
			CameraHeight = -HIGHEST_CAMERA_Y;
		}

		if((pad & PADL1) && (CameraDistance > CLOSEST_CAMERA_DISTANCE)) {
			CameraDistance -= 40;
		}
		if((pad & PADL2) && (CameraDistance < FURTHEST_CAMERA_DISTANCE)) {
			CameraDistance += 40;
		}

			// Calculate the other viewing variables given the new height
			// and distance.

		CameraLookUpDownAngle = ratan2(-CameraHeight - CAMERA_LEVEL_Y,CameraDistance);

		CameraRotate.vy = (-CameraAngle + ONE / 2);
		CameraRotate.vx = CameraLookUpDownAngle;
		CameraRotate.vz = 0;
		TopView.vpx = rsin(CameraAngle) * CameraDistance / ONE;
		TopView.vpz = rcos(CameraAngle) * CameraDistance / ONE;

			// React to key presses.

		if((pad & PADR3) && (mouthMoveFrame == 0)) {
				// Only roar if the dino isn't currently roaring.
			doRoarSound();
			mouthMoveFrame = 16;
		}

		if((pad & PADRright) && (tailAnimFrame <= 0)) {
				// Wag tail if the tail isn't currently wagging.
			tailAnimFrame = 1024;
		}

		if(pad & PADRup) {
			quit = TRUE;
		}

				// Counts down timers, rumbles the display and animates
				// the volcano texture within the environment models.
		processVolcano();

		if((pad & PADL3) && StompStatus == FALSE) {	// not currently doing a stomp
			desiredState = STOMP;
			StompStatus = TRUE;	// Doing a stomp
		}

		if( (pad & (PADR1 | PADR2)) && (StompStatus == FALSE)) {
				// If they are pressing either of the 'walk' buttons, and
				// we're not doing a stomp...
			desiredState = (pad & PADR1)?WALK1_TO_WALK2:BACKWALK1_TO_BACKWALK2;
			WalkTriceratops(TRUE);
		} else {

			if(StompStatus == FALSE) {
				desiredState = STOPPED;
			}
			if(BoredTimer == 0) {
				desiredState = BORED;
				BoredTimer = 400 + rand() % 500;
			}
			WalkTriceratops(FALSE);
		}

			// Reduce the "Bored" Timer. The dino shifts his weight around
			// when he's bored.
		if(currentState == STOPPED && BoredTimer > 0) {
			BoredTimer--;
		}

		processSounds();

		TopView.vpx += CameraShudder_LR;
		TopView.vpy = CameraHeight + CameraShudder_UD;

			// move the sky u/v's so that it looks 3D.
		moveSky(TopView.vpx,TopView.vpz);

			// Create the scenes.
		CreateOTforReflection();
		CreateOTforGround();
		CreateOTforMainScene();

#ifdef MESSAGES_ON
		FntPrint("Camera Pos: %d %d %d\n",TopView.vpx,TopView.vpy,TopView.vpz);
		FntPrint("Camera Dist: %d\n",CameraDistance);
		FntPrint("Camera LookUpDownAngle: %d\n",CameraLookUpDownAngle);


		FntFlush(-1);
#endif

		processController();

		completeDraw();				// Wait for GPU to finish then flip screen.

		oldpad = pad;
	}

	exitProgram();
	return 0;
}

